/* Debug Header */
#if DEBUG
# define IFDEBUG(x) do { if (debug) x; }while(0)
# define IFDEBUG5(x) do { if (debug) x; }while(0)
#else
# define IFDEBUG(x)
# define IFDEBUG5(x)
#endif
/* End Debug Header */


#if HAVE_CONFIG_H
# include <config.h>
#endif


#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <pthread.h>

#include "defines.h"
#include "dbconfig.h"
#include "extern.h"

/* declared in dbconfig.h
 * int debug=0;
 * int verbose=0;
 * struct dbreq_struct req[MAX_REQUESTS];
 * struct conn_struct connection[MAX_REQUESTS];
 * struct dbreq_struct req[MAX_REQUESTS];
 */

int request[MAX_REQUESTS];
int thread_started;

struct conn_struct connection[MAX_REQUESTS];
struct dbreq_struct req[MAX_REQUESTS];
int debug=0;
int verbose=0;

struct dbconf_struct *config;
struct dbcon_struct *db;

int fork_db_thread(void);
void open_dbs(void);
void openlogs(void);


int main(int argc, char **argv) {

	int i;
	int r;
	int f;
	int socketd;

	f=0;

	if (parse_config(argc, argv)) {
		fprintf(stderr, "Unable to parse config file, exiting...\n");
		exit(-1);
	}
	setup_sighandler();
	f=fork();
	if (!f) {
		/* child */
		sleep(1); /* sleep 1 second to parent process to close files and exit */
	} else {
		/* parent (main) thread */
		// close_local_dbs(); /* parent doesn't need to access the db */
		cleanup();
		if (f<0) {
			fprintf(stderr, "Fatal: Unable to fork()\n");
			exit(1);
		}
		fprintf(stdout, "Running Background [%i]\n", f);
		exit(0);
	}
	
	/* we are the child here */

	openlogs();
	open_dbs();
	start_threads();
	socketd=setup_port();
	start_listener(socketd);
	
	exit(10);

}

void openlogs(void) {

	if (!(config->logfile && config->errfile)) {
		fprintf(stderr, "You must define logfiles!\n");
		cleanup();
		exit(1);
	}
	config->log=freopen(config->logfile, "w+", stdout);
	if (!config->log) {
		perror("Error opening logfile");

	}
	config->err=freopen(config->errfile, "w+", stderr);
	if (!config->err) {
		perror("Error opeing error log");
	}

	if (!config->log || !config->err) {
		cleanup();
		exit(1);
	}

}

/* Lots of ifdefs in the next functions....
 *
 * whilst this is ugly, I have done it so at compile
 * time the user can compile for the databases they
 * may want to use.
 *
 * This will then allow at runtime for the user to
 * select the Database type they want to use, without
 * symbol errors, or a recompile etc...
 *
 * Now if anyone knows a better way to do this I am
 * happy to learn ;-)
 *
 */
void open_dbs(void) {

	char errmsg[50];
	int dbfail=0;

	switch (config->dbtype) {
		case LOCAL_DB:
#ifdef HAVE_FILES
			open_local_dbs();
#else
			dbfail++;
			sprintf(errmsg, "the local database");
#endif
			break;
		case MYSQL_DB:
#ifdef HAVE_MYSQL
			open_mysql_db();
#else
			dbfail++;
			sprintf(errmsg, "MySQL databases");
#endif
			break;
		case PGSQL_DB:
#ifdef HAVE_PGSQL
			open_pgsql_db();
#else
			dbfail++;
			sprintf(errmsg, "Postgres databases");
#endif
		case MSQL_DB:
#ifdef HAVE_MSQL
			open_msql_db();
#else
			dbfail++;
			sprintf(errmsg, "mSQL databases");
#endif
			break;
		case IODBC_DB:
#ifdef HAVE_IODBC
			open_iodbc_db();
#else
			dbfail++;
			sprintf(errmsg, "iODBC databases");
#endif
			break;
		case UNIXODBC_DB:
#ifdef HAVE_UNIXODBC
			open_unixodbc_db();
#else
			dbfail++;
			sprintf(errmsg, "Postgres databases");
#endif
			break;
		case DB2_DB:
#ifdef HAVE_DB2
			open_db2_db();
#else
			dbfail++;
			sprintf(errmsg, "IBM DB2 databases");
#endif
			break;
		case SOLID_DB:
#ifdef HAVE_SOLID
			open_solid_db();
#else
			dbfail++;
			sprintf(errmsg, "Solid SQL databases");
#endif
			break;
		case OPENLINK_DB:
#ifdef HAVE_VIRT
			open_openlink_db();
#else
			dbfail++;
			sprintf(errmsg, "Openlink ODBC (Virtuoso) databases");
#endif
			break;
		case EASYSOFT_DB:
#ifdef HAVE_EASYSOFT
			open_easysoft_db();
#else
			dbfail++;
			sprintf(errmsg, "EasySoft ODBC databases");
#endif
			break;
		case SAPDB_DB:
#ifdef HAVE_SAPDB
			open_sap_db();
#else
			dbfail++;
			sprintf(errmsg, "SAP ODBC databases");
#endif
			break;
		case IBASE_DB:
#ifdef HAVE_IBASE
			open_ibase_db();
#else
			dbfail++;
			sprintf(errmsg, "InterBase SQL databases");
#endif
			break;
		case CTLIB_DB:
#ifdef HAVE_CTLIB
			open_ctlib_db();
#else
			dbfail++;
			sprintf(errmsg, "CT-LIB SyBase SQL databases");
#endif
			break;
		case ORACLE8_DB:
#ifdef HAVE_ORACLE8
			open_oracle8_db();
#else
			dbfail++;
			sprintf(errmsg, "Oracle Version 8 databases");
#endif
			break;
		case ORACLE7_DB:
#ifdef HAVE_ORACLE7
			open_oracle7_db();
#else
			dbfail++;
			sprintf(errmsg, "Oracle Version 7 databases");
#endif
			break;
		default:
			fprintf(stderr, "ERROR: No Databases defined, exiting!\n");
			exit(1);
			break;
	}
	if (dbfail) {
		fprintf(stderr, "ERROR: support for %s is not compiled in!\n", errmsg);
		exit(1);
	}

}

int do_dbadd(int id) {

	switch (config->dbtype) {
		case LOCAL_DB:
#ifdef HAVE_FILES
			return(add_local_dbs(id));
#endif
			break;
		case MYSQL_DB:
#ifdef HAVE_MYSQL
			return(add_mysql_dbs());
#endif
			break;
		case PGSQL_DB:
#ifdef HAVE_PGSQL
			return(add_pgsql_dbs());
#endif
		case MSQL_DB:
#ifdef HAVE_MSQL
			return(add_msql_dbs());
#endif
			break;
		case IODBC_DB:
#ifdef HAVE_IODBC
			return(add_iodbc_dbs());
#endif
			break;
		case UNIXODBC_DB:
#ifdef HAVE_UNIXODBC
			return(add_unixodbc_dbs());
#endif
			break;
		case DB2_DB:
#ifdef HAVE_DB2
			return(add_db2_dbs());
#endif
			break;
		case SOLID_DB:
#ifdef HAVE_SOLID
			return(add_solid_dbs());
#endif
			break;
		case OPENLINK_DB:
#ifdef HAVE_VIRT
			return(add_openlink_dbs());
#endif
			break;
		case EASYSOFT_DB:
#ifdef HAVE_EASYSOFT
			return(add_easysoft_dbs());
#endif
			break;
		case SAPDB_DB:
#ifdef HAVE_SAPDB
			return(add_sap_dbs());
#endif
			break;
		case IBASE_DB:
#ifdef HAVE_IBASE
			return(add_ibase_dbs());
#endif
			break;
		case CTLIB_DB:
#ifdef HAVE_CTLIB
			return(add_ctlib_dbs());
#endif
			break;
		case ORACLE8_DB:
#ifdef HAVE_ORACLE8
			return(add_oracle8_dbs());
#endif
			break;
		case ORACLE7_DB:
#ifdef HAVE_ORACLE7
			return(add_oracle7_dbs());
#endif
			break;
		default:
			return(-1);
			break;
	}
}

int do_dbdel(int id) {

	switch (config->dbtype) {
		case LOCAL_DB:
#ifdef HAVE_FILES
			return(del_local_dbs());
#endif
			break;
		case MYSQL_DB:
#ifdef HAVE_MYSQL
			return(del_mysql_dbs());
#endif
			break;
		case PGSQL_DB:
#ifdef HAVE_PGSQL
			return(del_pgsql_dbs());
#endif
		case MSQL_DB:
#ifdef HAVE_MSQL
			return(del_msql_dbs());
#endif
			break;
		case IODBC_DB:
#ifdef HAVE_IODBC
			return(del_iodbc_dbs());
#endif
			break;
		case UNIXODBC_DB:
#ifdef HAVE_UNIXODBC
			return(del_unixodbc_dbs());
#endif
			break;
		case DB2_DB:
#ifdef HAVE_DB2
			return(del_db2_dbs());
#endif
			break;
		case SOLID_DB:
#ifdef HAVE_SOLID
			return(del_solid_dbs());
#endif
			break;
		case OPENLINK_DB:
#ifdef HAVE_VIRT
			return(del_dellink_dbs());
#endif
			break;
		case EASYSOFT_DB:
#ifdef HAVE_EASYSOFT
			return(del_easysoft_dbs());
#endif
			break;
		case SAPDB_DB:
#ifdef HAVE_SAPDB
			return(del_sap_dbs());
#endif
			break;
		case IBASE_DB:
#ifdef HAVE_IBASE
			return(del_ibase_dbs());
#endif
			break;
		case CTLIB_DB:
#ifdef HAVE_CTLIB
			return(del_ctlib_dbs());
#endif
			break;
		case ORACLE8_DB:
#ifdef HAVE_ORACLE8
			return(del_oracle8_dbs());
#endif
			break;
		case ORACLE7_DB:
#ifdef HAVE_ORACLE7
			return(del_oracle7_dbs());
#endif
			break;
		default:
			return(-1);
			break;
	}
}

int do_dbcheck(int id) {

	switch (config->dbtype) {
		case LOCAL_DB:
#ifdef HAVE_FILES
			return(check_local_dbs(id));
#endif
			break;
		case MYSQL_DB:
#ifdef HAVE_MYSQL
			return(check_mysql_dbs(id));
#endif
			break;
		case PGSQL_DB:
#ifdef HAVE_PGSQL
			return(check_pgsql_dbs(id));
#endif
		case MSQL_DB:
#ifdef HAVE_MSQL
			return(check_msql_dbs(id));
#endif
			break;
		case IODBC_DB:
#ifdef HAVE_IODBC
			return(check_iodbc_dbs(id));
#endif
			break;
		case UNIXODBC_DB:
#ifdef HAVE_UNIXODBC
			return(check_unixodbc_dbs(id));
#endif
			break;
		case DB2_DB:
#ifdef HAVE_DB2
			return(check_db2_dbs(id));
#endif
			break;
		case SOLID_DB:
#ifdef HAVE_SOLID
			return(check_solid_dbs(id));
#endif
			break;
		case OPENLINK_DB:
#ifdef HAVE_VIRT
			return(check_openlink_dbs(id));
#endif
			break;
		case EASYSOFT_DB:
#ifdef HAVE_EASYSOFT
			return(check_easysoft_dbs(id));
#endif
			break;
		case SAPDB_DB:
#ifdef HAVE_SAPDB
			return(check_sap_dbs(id));
#endif
			break;
		case IBASE_DB:
#ifdef HAVE_IBASE
			return(check_ibase_dbs(id));
#endif
			break;
		case CTLIB_DB:
#ifdef HAVE_CTLIB
			return(check_ctlib_dbs(id));
#endif
			break;
		case ORACLE8_DB:
#ifdef HAVE_ORACLE8
			return(check_oracle8_dbs(id));
#endif
			break;
		case ORACLE7_DB:
#ifdef HAVE_ORACLE7
			return(check_oracle7_dbs(id));
#endif
			break;
		default:
			return -1;
			break;

	}
	/* keep for example - Mat
		case LOCAL_DB:
			return(check_local_dbs(id));
			break;
		case MYSQL_DB:
			return(check_mysql_dbs(connection[id]));
			break;
			*/
}

void init_db_threads () {

	switch (config->dbtype) {
		case LOCAL_DB:
			break;
		case MYSQL_DB:
#ifdef HAVE_THREADSAFE_MYSQL
			init_mysql_threads();
#endif
			break;
		case PGSQL_DB:
#ifdef HAVE_PGSQL
			init_pgsql_threads();
#endif
		case MSQL_DB:
#ifdef HAVE_MSQL
			init_msql_threads();
#endif
			break;
		case IODBC_DB:
#ifdef HAVE_IODBC
			init_iodbc_threads();
#endif
			break;
		case UNIXODBC_DB:
#ifdef HAVE_UNIXODBC
			init_unixodbc_threads();
#endif
			break;
		case DB2_DB:
#ifdef HAVE_DB2
			init_db2_threads();
#endif
			break;
		case SOLID_DB:
#ifdef HAVE_SOLID
			init_solid_threads();
#endif
			break;
		case OPENLINK_DB:
#ifdef HAVE_VIRT
			init_virt_threads();
#endif
			break;
		case EASYSOFT_DB:
#ifdef HAVE_EASYSOFT
			init_easysoft_threads();
#endif
			break;
		case SAPDB_DB:
#ifdef HAVE_SAPDB
			init_sap_threads();
#endif
			break;
		case IBASE_DB:
#ifdef HAVE_IBASE
			init_ibase_threads();
#endif
			break;
		case CTLIB_DB:
#ifdef HAVE_CTLIB
			init_ctlib_threads();
#endif
			break;
		case ORACLE8_DB:
#ifdef HAVE_ORACLE8
			init_oracle8_threads();
#endif
			break;
		case ORACLE7_DB:
#ifdef HAVE_ORACLE7
			init_oracle7_threads();
#endif
			break;
		default:
			break;
	}
}
