/* Debug header */
#if DEBUG
# define IFDEBUG(x) do { if (debug) x; }while(0);
# define IFDEBUG5(x) do { if (debug==5) x; }while(0);
#else
# define IFDEBUG(x)
# define IFDEBUG5(x)
#endif
/* End Debug header */

#if HAVE_CONFIG_H
# include <config.h>
#endif


#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "extern.h"
#include "defines.h"
#include "dbconfig.h"
#include "funcs.h"

#ifdef HAVE_PTHREAD
# include <pthread.h>
#else
# error Must have pthread support to compile with the built in DB!
#endif

/* locks for locking the db's before writing to them */
pthread_mutex_t http_lock=PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t smtp_lock=PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t socks_lock=PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t trap_lock=PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t white_lock=PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t local_lock=PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t oklist_lock=PTHREAD_MUTEX_INITIALIZER;

void get_file_lock(int);
void drop_file_lock(int);
void sanitise_data(int, char *);

void open_local_dbs(void) {

	int r=0;

	dbinfo=(struct db_struct *)calloc(sizeof(struct db_struct), 1);

	IFDEBUG(print_debug(5, "Opening SpamTrap users list file (%s)\n", config->trap));
	if (open_local_db(&dbinfo->trapfd, config->trap)) r=1;
	IFDEBUG(print_debug(5, "Opening bad Reverse DNS list file (%s)\n", config->rdns));
	if (open_local_db(&dbinfo->rdnsfd, config->rdns)) r=1;
	IFDEBUG(print_debug(5, "Opening open-relay list file (%s)\n", config->smtp));
	if (open_local_db(&dbinfo->relayfd, config->smtp)) r=1;
	IFDEBUG(print_debug(5, "Opening http proxy list file (%s)\n", config->http));
	if (open_local_db(&dbinfo->proxyfd, config->http)) r=1;
	IFDEBUG(print_debug(5, "Opening socks list file (%s)\n", config->sock));
	if (open_local_db(&dbinfo->socksfd, config->sock)) r=1;
	IFDEBUG(print_debug(5, "Opening white list file (%s)\n", config->white));
	if (open_local_db(&dbinfo->whitefd, config->white)) r=1;
	IFDEBUG(print_debug(5, "Opening local_db file (%s)\n", config->oklist));
	if (open_local_db(&dbinfo->oklistfd, config->oklist)) r=1;

	if (r) {
		cleanup();
		exit(1);
	}

}

void close_local_dbs(void) {

	close(dbinfo->trapfd);
	close(dbinfo->rdnsfd);
	close(dbinfo->relayfd);
	close(dbinfo->proxyfd);
	close(dbinfo->socksfd);
	close(dbinfo->whitefd);
	close(dbinfo->oklistfd);

	IFDEBUG(print_debug(5, "Freeing & Clearing 'dbinfo'\n"));
	free(dbinfo);
	dbinfo=NULL;
}

int open_local_db(int *fd, char *filename) {

	*fd=open(filename, O_CREAT | O_RDWR, S_IRUSR | S_IWUSR);
	if (*fd<0) {
		fprintf(stderr, "[%s] ", filename);
		perror("Error opening file");
		return ERROR;
	}
	return 0;

}
int reopen_local_db(int *fd, char *filename) {

	IFDEBUG(print_debug(5, "Closing %i\n", *fd));
	close(*fd);

	IFDEBUG(print_debug(5, "Opening %s\n", filename));
	*fd=open(filename, O_CREAT | O_TRUNC | O_RDWR, S_IRUSR | S_IWUSR);
	if (*fd<0) {
		fprintf(stderr, "[%s] ", filename);
		perror("Error opening file");
		return ERROR;
	}

	IFDEBUG(print_debug(5, "New fd=%i returning 0\n", *fd));
	return 0;

}

int reopen_dbs(int *fd, int type) {

	int r=0;

	switch(type) {
		case TRAP:
			IFDEBUG(print_debug(5, "Re-Opening SpamTrap users list file (%s)\n", config->trap));
			if (reopen_local_db(&dbinfo->trapfd, config->trap)) r=1;
			*fd=dbinfo->trapfd;
			break;
		case RDNS:
			IFDEBUG(print_debug(5, "Re-Opening bad Reverse DNS list file (%s)\n", config->rdns));
			if (reopen_local_db(&dbinfo->rdnsfd, config->rdns)) r=1;
			*fd=dbinfo->rdnsfd;
			break;
		case SMTP:
			IFDEBUG(print_debug(5, "Re-Opening open-relay list file (%s)\n", config->smtp));
			if (reopen_local_db(&dbinfo->relayfd, config->smtp)) r=1;
			*fd=dbinfo->relayfd;
			break;
		case HTTP:
			IFDEBUG(print_debug(5, "Re-Opening http proxy list file (%s)\n", config->http));
			if (reopen_local_db(&dbinfo->proxyfd, config->http)) r=1;
			*fd=dbinfo->proxyfd;
			break;
		case SOCKS:
			IFDEBUG(print_debug(5, "Re-Opening socks list file (%s)\n", config->sock));
			if (reopen_local_db(&dbinfo->socksfd, config->sock)) r=1;
			*fd=dbinfo->socksfd;
			break;
		case WHITE:
			IFDEBUG(print_debug(5, "Re-Opening white list file (%s)\n", config->white));
			if (reopen_local_db(&dbinfo->whitefd, config->white)) r=1;
			*fd=dbinfo->whitefd;
			break;
		case OKLIST:
			IFDEBUG(print_debug(5, "Re-Opening local_db file (%s)\n", config->oklist));
			if (reopen_local_db(&dbinfo->oklistfd, config->oklist)) r=1;
			*fd=dbinfo->oklistfd;
			break;
		default:
			r++;
			break;
	}
	if (r) {
		fprintf(stderr, "reopen_db(): Arrrgghh! Error reopening file exiting!\n");
		cleanup();
		exit(1);
	}

}

int rewrite_file(int fd, int type, int blen, char *buf) {

	int len=0;
	
	reopen_dbs(&fd, type);

	/* reset pointer to beginning of the file */
	lseek(fd, 0, SEEK_SET);

	IFDEBUG(print_debug(5, "Writing %i bytes to disk\n", blen));

	len=write(fd, buf, blen);
	if (len !=  blen) {
		IFDEBUG(print_debug(5, "Error write() returned: %i, should have been %i\n", len, blen));
		if (len < 0) perror("Error writing to disk");
		return DELPROB;
	}

	if (fsync(fd)) {
		perror("Error syncing to disk");
		return DELPROB;
	}
	return DELETED;	
}

int add_local_dbs(int id) {

	int fd=0;
	int ret=0;
	
	IFDEBUG(print_debug(5, "Getting lock for DB add...\n"));
	get_file_lock(req[id].client->query);
	switch(req[id].client->query) {
		case HTTP:
			IFDEBUG(print_debug(2, "Thread[%i]: adding to the HTTP list\n", id));
			fd=dbinfo->proxyfd;
			break;
		case SMTP:
			IFDEBUG(print_debug(2, "Thread[%i]: adding to the SMTP list\n", id));
			fd=dbinfo->relayfd;
			break;
		case SOCKS:
			IFDEBUG(print_debug(2, "Thread[%i]: adding to the SOCKS list\n", id));
			fd=dbinfo->socksfd;
			break;
		case TRAP:
			IFDEBUG(print_debug(2, "Thread[%i]: adding to the TRAP list\n", id));
			fd=dbinfo->trapfd;
			break;
		case WHITE:
			IFDEBUG(print_debug(2, "Thread[%i]: adding to the WHITE list\n", id));
			fd=dbinfo->whitefd;
			break;
		case OKLIST:
			IFDEBUG(print_debug(2, "Thread[%i]: adding to the OK list\n", id));
			fd=dbinfo->oklistfd;
			break;
		case RDNS:
			IFDEBUG(print_debug(2, "Thread[%i]: adding to the RDNS list\n", id));
			fd=dbinfo->rdnsfd;
			break;
		default:
			IFDEBUG(print_debug(2, "Thread[%i]: Unknown list type %i\n", id, req[id].client->query));
			return ERROR;
	}
	return(add_to_file(fd, req[id].client->query, req[id].client->ipaddr));
}

int del_local_dbs(int id) {

	int fd=0;
	int ret=0;
	
	IFDEBUG(print_debug(5, "Getting lock for DB delete...\n"));
	get_file_lock(req[id].client->query);
	switch(req[id].client->query) {
		case HTTP:
			IFDEBUG(print_debug(2, "Thread[%i]: deleting from the HTTP list\n", id));
			fd=dbinfo->proxyfd;
			break;
		case SMTP:
			IFDEBUG(print_debug(2, "Thread[%i]: deleting from the SMTP list\n", id));
			fd=dbinfo->relayfd;
			break;
		case SOCKS:
			IFDEBUG(print_debug(2, "Thread[%i]: deleting from the SOCKS list\n", id));
			fd=dbinfo->socksfd;
			break;
		case TRAP:
			IFDEBUG(print_debug(2, "Thread[%i]: deleting from the TRAP list\n", id));
			fd=dbinfo->trapfd;
			break;
		case WHITE:
			IFDEBUG(print_debug(2, "Thread[%i]: deleting from the WHITE list\n", id));
			fd=dbinfo->whitefd;
			break;
		case OKLIST:
			IFDEBUG(print_debug(2, "Thread[%i]: deleting from the OK list\n", id));
			fd=dbinfo->oklistfd;
			break;
		case RDNS:
			IFDEBUG(print_debug(2, "Thread[%i]: deleting from the RDNS list\n", id));
			fd=dbinfo->rdnsfd;
			break;
		default:
			IFDEBUG(print_debug(2, "Thread[%i]: Unknown list type %i\n", id, req[id].client->query));
			return ERROR;
	}
	return(delete_from_file(fd, req[id].client->query, req[id].client->ipaddr));
}

int check_local_dbs(int id) {

	int fd=0;
	int ret=0;
	
	IFDEBUG(print_debug(5, "Getting lock for DB query...\n"));
	get_file_lock(req[id].client->query);
	switch(req[id].query) {
		case HTTP:
			IFDEBUG(print_debug(2, "Thread[%i]: Checking HTTP list\n", id));
			fd=dbinfo->proxyfd;
			break;
		case SMTP:
			IFDEBUG(print_debug(2, "Thread[%i]: Checking SMTP list\n", id));
			fd=dbinfo->relayfd;
			break;
		case SOCKS:
			IFDEBUG(print_debug(2, "Thread[%i]: Checking SOCKS list\n", id));
			fd=dbinfo->socksfd;
			break;
		case TRAP:
			IFDEBUG(print_debug(2, "Thread[%i]: Checking TRAP list\n", id));
			fd=dbinfo->trapfd;
			break;
		case WHITE:
			IFDEBUG(print_debug(2, "Thread[%i]: Checking WHITE list\n", id));
			fd=dbinfo->whitefd;
			break;
		case OKLIST:
			IFDEBUG(print_debug(2, "Thread[%i]: Checking OK list\n", id));
			fd=dbinfo->oklistfd;
			break;
		case RDNS:
			IFDEBUG(print_debug(2, "Thread[%i]: Checking RDNS list\n", id));
			fd=dbinfo->rdnsfd;
			break;
		default:
			IFDEBUG(print_debug(2, "Thread[%i]: Unknown list type %i\n", id));
			return ERROR;
	}
	
	ret=scan_file(fd, req[id].client->query, req[id].client->ipaddr);
	return ret;
}

int add_to_file(int fd, int type, char *ipaddr) {

	char buf[256];

	memset(buf, '\0', 256);
	sprintf(buf, "%s\n", ipaddr);

	lseek(fd, 0, SEEK_END);
	write(fd, buf, strlen(buf));
	fsync(fd); /* this is used instead of fdatasync() deliberately */
	drop_file_lock(type);
	return(ADDED);
}

int delete_from_file(int fd, int type, char *ipaddr) {

	int ret=0;
	char *file, *p, *buf;
	char *lt, *tok, delim[2] = { '\n', '\0' };
	int len=0, blen=0;

	char t[256] = {};

	IFDEBUG(print_debug(4, "Delete %s requested\n", ipaddr));
	
	ret=scan_file(fd, type, ipaddr);
	if (ret==NOT_MATCHED) return DELPROB;
	ret=DELPROB; /* reset value as this is our default function exit now */

	/* we have a match... now we need to delete it */

	lseek(fd, 0, SEEK_SET);
	
	file=p=read_file(fd);
	if (!file) goto problem;

	len=strlen(ipaddr);
	blen=strlen(file);
	blen++;
	buf=malloc(blen); /* we are going to copy the entire file into this memory to save on writes */
	memset(buf, '\0', blen);

	tok=GetToken(p, delim, &lt);
	while(tok) {
		IFDEBUG(print_debug(5, "Tok=%s\n", tok));
		p=tok;
		while (*p && (*p == '\t' || *p == ' ')) p++; /* remove all leading white spaces */
		memset(t, '\0', 256);
		sprintf(t, "%s\n", tok);

		if (!strncmp(p, ipaddr, len)) {
			IFDEBUG(print_debug(5, "Found %s in %s", ipaddr, t));
			/* we don't do anything with it as we need to nuke it */
		} else {
			IFDEBUG(print_debug(5, "Adding %s", t));
			strcat(buf, t); /* not a match so copy it into the new buffer */
		}
		IFDEBUG(print_debug(5, "buf is currently:\n%s", buf));
		tok=GetToken(NULL, delim, &lt);
	}

	sanitise_data(blen, buf);

	blen=strlen(buf); /* after the sanitise we may have a different length to write */

	/* ok we now have the entire file in 'buf' (without the 'deleted' address/line) */
	ret=rewrite_file(fd, type, blen, buf);

	/* cleanup and return */
problem:
	drop_file_lock(type);
	return ret;
}

void sanitise_data(int blen, char *buf) {

	char *p;
	int i;

	p=buf;

	for(i=0 ; i<blen; i++) {
		if (*buf) {
			*(p++)=*(buf++);
		} else {
			buf++;
		}
	}
	*p='\0'; /* terminate the string/file */

}

int scan_file(int fd, int type, char *ipaddr) {
	
	char *buf, *file;
	char *lt, *tok, delim[2] = { '\n', '\0' };
	char *line, *p, *t;
	int len;

	len=strlen(ipaddr);

	lseek(fd, 0, SEEK_SET);
	
	file=p=read_file(fd); /* file is the initial pointer */
	drop_file_lock(type);
	if (!file) return NOT_MATCHED;

	IFDEBUG(print_debug(4, "Searching for: %s\n", ipaddr));

	tok=GetToken(p, delim, &lt);
	while (tok && *tok != '#') {
	
		p=tok;
		while (*p && (*p == '\t' || *p == ' ')) p++;
		if (*p==*ipaddr) {
			IFDEBUG(print_debug(5, "Checking %i chararcter for match: %s & %s\n", len, p, ipaddr));
			if (!strncmp(p, ipaddr, len)) {
				IFDEBUG(print_debug(4, "Searched for: %s - matched!\n", ipaddr));
				/* we have a match! */
				free(file);
				return MATCHED;
			}
		}
		tok=GetToken(NULL, delim, &lt);
	}
	IFDEBUG(print_debug(4, "Searched for: %s - NOT matched!\n", ipaddr));
	free(file);
	return NOT_MATCHED;
}

void get_file_lock(int type) {

	switch (type) {
		case HTTP:
			pthread_mutex_lock(&http_lock);
			break;
		case SMTP:
			pthread_mutex_lock(&smtp_lock);
			break;
		case SOCKS:
			pthread_mutex_lock(&socks_lock);
			break;
		case TRAP:
			pthread_mutex_lock(&trap_lock);
			break;
		case WHITE:
			pthread_mutex_lock(&white_lock);
			break;
		case LOCALLIST:
			pthread_mutex_lock(&local_lock);
			break;
		case OKLIST:
			pthread_mutex_lock(&oklist_lock);
			break;
	}

}

void drop_file_lock(int type) {

	switch (type) {
		case HTTP:
			pthread_mutex_unlock(&http_lock);
			break;
		case SMTP:
			pthread_mutex_unlock(&smtp_lock);
			break;
		case SOCKS:
			pthread_mutex_unlock(&socks_lock);
			break;
		case TRAP:
			pthread_mutex_unlock(&trap_lock);
			break;
		case WHITE:
			pthread_mutex_unlock(&white_lock);
			break;
		case LOCALLIST:
			pthread_mutex_unlock(&local_lock);
			break;
		case OKLIST:
			pthread_mutex_unlock(&oklist_lock);
			break;
	}
}
