/* Debug Header */
#if DEBUG
# define IFDEBUG(x) do { if (debug) x; }while(0)
# define IFDEBUG5(x) do { if (debug) x; }while(0)
#else
# define IFDEBUG(x)
# define IFDEBUG5(x)
#endif
/* End Debug Header */

#if HAVE_CONFIG_H
# include <config.h>
#endif

/* This file will only be compiled if --with-mysql is specified on the
 * configure command line.
 *
 * This file contains all that is nessesary to talk to a MySQL server.
 */

#include <stdio.h>
#include <stdarg.h>


#include "dbconfig.h"
#include "extern.h"
#include "defines.h"
#include <mysql.h>

void connect_mysqldb();

int thread_safe = 0; /* this is used for indicating whether the MySQL DB client is threadsafe. */

MYSQL mysql; /* this holds all the connection info for the connection the the backend DB. */
MYSQL_RES *result;
MYSQL_ROW row;

void open_mysql_db(void) {

#ifdef HAVE_THREADSAFE_MYSQL
	IFDEBUG(print_debug(5, "Checking for thread safe MySQL client\n"));
	thread_safe=mysql_thread_safe();
	if (thread_safe) my_init();
#endif

	IFDEBUG(print_debug(5, "Initialising MySQL connection structure\n"));
	mysql_init(&mysql);

	IFDEBUG(print_debug(5, "Making MySQL connection\n"));
	connect_mysqldb();

}

void connect_mysqldb () {
	
	unsigned int client_flag=0;

	/* Defaults for the MySQL connection are as follows:
	 *
	 * Database Host: localhost
	 * Database Port: 3306
	 * Database Name: sorbs
	 * Username: sorbs
	 * Password: 
	 * Unix Socket and path: /tmp/mysql.sock
	 */

	IFDEBUG(print_debug(5, "Attempting to connect to MySQL DB using: mysql://%s:<not-shown>@%s:%i/%s/\n", db->user?db->user:"sorbs", db->host?db->host:"localhost", db->port?db->port:3306, db->name?db->name:"sorbs"));

	mysql_options(&mysql, MYSQL_OPT_COMPRESS, 0);
	IFDEBUG(print_debug(5, "Set Options....%p\n", &mysql));
	
	if (!mysql_real_connect(&mysql,
				db->host?db->host:"localhost",
				db->user?db->user:"sorbs",
				db->pass?db->pass:"password",
				db->name?db->name:"sorbs",
				db->port?db->port:3306,
				db->unix_socket?db->unix_socket:"/tmp/mysql.sock",
				client_flag)) {
		fprintf(stderr, "Failed to connect to database: Error: %s\n", mysql_error(&mysql));
		cleanup();
		exit(1);
	} else {
		IFDEBUG(print_debug(3, "Connected to MySQL DB using: mysql://%s:<not-shown>@%s:%i/%s/\n", db->user?db->user:"sorbs", db->host?db->host:"localhost", db->port?db->port:3306, db->name?db->name:"sorbs"));
	}
}

void disconnect_db_thread() {

#ifdef HAVE_THREADSAFE_MYSQL
	if (thread_safe) {
		my_thread_end();
	}
#endif

}

int disconnect_db () {

	mysql_close(&mysql);

}

int query_result () {

	result=mysql_use_result(&mysql);
	while((row=mysql_fetch_row(result))) {
		// do something
	}
	if (!mysql_eof(result)) {
		fprintf(stderr, "Error: %s\n", mysql_error(&mysql));
	}
}

int add_mysql_dbs() {

	
	return(ADDPROB);
}

int del_mysql_dbs() {

	return(DELPROB);
}

int check_mysql_dbs(int id) {

	char *query;

	if (!&mysql) return(MATCHED);

	return NOT_MATCHED;

}

#ifdef HAVE_THREADSAFE_MYSQL
void init_mysql_threads() {

	my_thread_init();

}
#endif
