/* Debug Header */
#if DEBUG
# define IFDEBUG(x) do { if (debug) x; }while(0)
# define IFDEBUG5(x) do { if (debug) x; }while(0)
#else
# define IFDEBUG(x)
# define IFDEBUG5(x)
#endif
/* End Debug Header */

#if HAVE_CONFIG_H
# include <config.h>
#endif

/*
 * This file build all the SQL queries for all databases.
 */


#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>

#include "defines.h"
#include "dbconfig.h"
#include "extern.h"


char *vbuild_query (const char *fmt, va_list args) {

	int d, ret;
	char c, *p, *s;
	char *t, *str, *memp, *buf;
	char *tmp;

	buf=malloc(1024);
	tmp=malloc(512);

	memset(buf, '\0', 1024);
	memset(tmp, '\0', 512);

	memp=str=(char *)strdup(fmt);
	while(*str) {
		t=str;
		IFDEBUG(print_debug(5, "t=%s\n", t));
		t=(char *)strchr(t, '%');
		if (!t) break;
		*t='\0';
		if (str) strcat(buf, str);
		t++;
		str=t;
		*tmp='\0';
		switch(*str++) {
			case '%':
				sprintf(tmp, "%s%%", buf);
				IFDEBUG(print_debug(5, "tmp[.]=%s\n", tmp));
				break;
			case 's':
				s = va_arg(args, char *);
				sprintf(tmp, "%s%s", buf, s);
				IFDEBUG(print_debug(5, "tmp[s]=%s\n", tmp));
				break;
			case 'i':
				d = va_arg(args, int);
				sprintf(tmp, "%s%i", buf, d);
				IFDEBUG(print_debug(5, "tmp[i]=%s\n", tmp));
				break;
			case 'c':
				c = (char) va_arg(args, int);
				sprintf(tmp, "%s%c", buf, c);
				IFDEBUG(print_debug(5, "tmp[c]=%s\n", tmp));
				break;
			case 'p':
				s = va_arg(args, char *);
				sprintf(tmp, "%s%p", buf, s);
				IFDEBUG(print_debug(5, "tmp[p]=%s\n", tmp));
				break;
		}
		if (*tmp) strcpy(buf, tmp);
		IFDEBUG(print_debug(5, "buf=%s\n", buf));
	}
	if (str) strcat(buf, str);

	IFDEBUG(print_debug(4, "Final Query [%s]\n", buf));
	free(memp);
	free(tmp);
	return buf;
}

char *build_query (const char *fmt, ...) {

	char *r;
	va_list va;

	va_start(va, fmt);
	r=vbuild_query(fmt, va);
	va_end(va);

	return r;
}

char *create_query(int type, ...) {

	char *t;
	va_list args;
	static char *sql[] = {
/* 0 */			"SELECT * FROM %s WHERE address = '%s'",										/* SELECT_ALL_BY_ADDRESS */
/* 1 */			"INSERT INTO %s VALUES (%s)",																/* INSERT_VALUES */
/* 2 */			"UPDATE %s SET %s WHERE %s",																/* UPDATE_SET_WHERE */
/* 3 */			"SELECT * FROM %s WHERE ipaddr = '%s'",											/* SELECT_ALL_BY_IP */
/* 4 */			"SELECT * FROM %s WHERE ipaddr = '%s' AND address = '%s'",	/* SELECT_ALL_BY_IP_AND_ADDRESS */
/* 5 */			"SELECT %s FROM %s",																				/* SELECT_FROM */
/* 6 */			"SELECT %s FROM %s WHERE %s LIKE '%s'",											/* SELECT_FROM_WHERE_LIKE */
/* 7 */			"SELECT %s FROM %s WHERE %s LIMIT %i"												/* SELECT_DROM_WHERE_LIMIT */
			};
	
	va_start(args, type);
	t=build_query(sql[type], args);
	va_end(args);

	return t;
}
