/* Debug Header */
#if DEBUG
# define IFDEBUG(x) do { if (debug) x; }while(0)
# define IFDEBUG5(x) do { if (debug) x; }while(0)
#else
# define IFDEBUG(x)
# define IFDEBUG5(x)
#endif
/* End Debug Header */


#if HAVE_CONFIG_H
# include <config.h>
#endif

#include "config.h"

#include <sys/file.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>

#include "defines.h"
#include "extern.h"
#include "configuration.h"

int spin_lock(const char *filename) {

	int fd;

	IFDEBUG(print_debug(5, "Opening Lockfile.\n"));
	fd=open(filename, O_CREAT | O_RDWR | O_NONBLOCK, S_IRUSR | S_IWUSR);
	if (fd<0) {
		perror("Error opening lock/semaphore file");
		cleanup();
		exit(1);
	}
	while((lockf(fd, F_TEST, 0)<0)) {
		IFDEBUG(print_debug(5, "Spinning waiting for lock\n"));
		sleep(5);
	}
	IFDEBUG(print_debug(5, "Getting Lock.\n"));
	lockf(fd, F_LOCK, 0);
	return fd;
}

void release_lock(int fd) {

	IFDEBUG(print_debug(5, "Releasing Lock.\n"));
	lockf(fd, F_ULOCK, 0);
	IFDEBUG(print_debug(5, "Closing lockfile.\n"));
	close(fd);

}

