/* Debug Header */
#if DEBUG
# define IFDEBUG(x) do { if (debug) x; }while(0)
# define IFDEBUG5(x) do { if (debug) x; }while(0)
#else
# define IFDEBUG(x)
# define IFDEBUG5(x)
#endif
/* End Debug Header */


#if HAVE_CONFIG_H
# include <config.h>
#endif

#include <strings.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netdb.h>
#include <pwd.h>
#include <sys/time.h>
#include "configuration.h"
#include "extern.h"
#include "defines.h"

int test_proxy(const char *hostname, const int port, const char *testhost, const int testport) {

	int sock=0, open=0;
	int t, r, ver=4, status=0;
	int *verp, *statusp;
	char *msg;
	char *res, *resp;
	char junk[20], junk2[20], junk3[20];

	verp=&ver;
	statusp=&status;

	sock=net_connect(hostname, port);

	if (!sock) {
		return 0;
	}

	/* Paranoid mode */
	if (sock && !mode) return 1;

	/* Initial proxy check */
	msg=malloc(NET_BUFFER);
	memset(msg, '\0', NET_BUFFER);

	sprintf(msg, "CONNECT %s:%i HTTP/1.0\n\n", testhost, testport?testport:25);
	net_send(sock, msg);
	resp=res=net_read(sock);

	IFDEBUG(print_debug(3, "HTTP-PROXY Response:\n%s", res));
	sscanf(res, "%s %i %s", junk, statusp, junk2);
	if (status==200) {
		open=1;
		if (config->spamtrapuser && config->spamtrapuser[0]) do_send_mail(sock, strdup(junk2));
		goto ret;
	}
	if (status==407 && mode) {
		IFDEBUG(print_debug(1, "Response: %i (Auth-Required - HTTP-Proxy's should deny by default) - Rejecting...\n", status));
		open=1;
		goto ret;
	}
	open=0;
ret:
	free(msg);
	free(resp);
	return open;
}

int test_wingate(const char *hostname, const int port, const char *testhost, const int testport) {

	int sock=0, open=0;
	int t, r, ver=4, status=0;
	int *verp, *statusp;
	char *msg;
	char *res, *resp;
	char junk[20], junk2[20], junk3[20];

	verp=&ver;
	statusp=&status;

	sock=net_connect(hostname, port);

	if (!sock) {
		return 0;
	}

	/* Paranoid mode */
	if (sock && !mode) return 1;

	/* Initial proxy check */
	msg=malloc(NET_BUFFER);
	memset(msg, '\0', NET_BUFFER);

	sprintf(msg, "%s:%i\n\n", testhost, testport?testport:25);
	net_send(sock, msg);
	resp=res=net_read(sock);

	IFDEBUG(print_debug(3, "WinGate Response:\n%s", res));
	if (*res=='2') {
		open=1;
		if (config->spamtrapuser && config->spamtrapuser[0]) do_send_mail(sock, strdup(junk2));
		goto ret;
	}
	if (status==407 && mode) {
		IFDEBUG(print_debug(1, "Response: %i (Auth-Required - HTTP-Proxy's should deny by default) - Rejecting...\n", status));
		open=1;
		goto ret;
	}
	open=0;
ret:
	free(msg);
	free(resp);
	return open;
}

int test_sock(const char *hostname, const int port) {

	int sock=0, open=0;
	int t, r, ver=4, i;
	int *verp;
	char *msg;
	char *res, *resp;

	verp=&ver;

	/* Initial Socks check */
	msg=calloc(sizeof(char), NET_BUFFER);
	sock=net_connect(hostname, port);

	if (!sock) return 0;

	if (sock && !mode) { /* Paranoid mode */
		open=1;
		return open;
	}

	initial_request(sock, msg, verp); /* Create initial version request packet */
	close(sock); /* We're just checking for a v4/v5 server at first */

	switch (ver) { 
		case 0:
			/* Possible sockv4 server */
			IFDEBUG(print_debug(1, "Sending socks4 connection request\n"));
			ver=4;
			open=make_socks4_request(hostname, hostname, port, msg, verp, port);
			if (open) break;
			open=make_socks4_request("localhost", hostname, port, msg, verp, port);
			if (open) break;
			/* FIXME */
			/* check user supplied hosts and ports for connection */

			if (config->testhost) {
				for (i=0;config->testhost[i] && !open ;i++) {
					open=make_socks4_request(config->testhost[i]->hostname, hostname, port, msg, verp, config->testhost[i]->port); 
				}
			}
			break;
		case 5:
			IFDEBUG(print_debug(1, "Sending socks5 connection request\n"));
			sock=net_connect(hostname, port);
			r=initial_request(sock, msg, verp); /* Create method request packet */
			close(sock); /* We're just checking for a server at first */
			if (r) {
				IFDEBUG(print_debug(0, "%s Authentication Required by server\n", r==2?"User/Pass":"GSAPI"));
			} else {
				if (mode==1) {
					open=1;
					IFDEBUG(print_debug(1, "Normal Mode - Rejecting Socks server....\n"));
				} else {
					IFDEBUG(print_debug(1, "Sending connect request with no authentication....\n"));
				}
			}
			if (open) break;

			open=make_socks5_request(hostname, hostname, port, msg, verp, port);
			if (open) break;

			/* Always check a sock proxy for connection to itself, cause you can stack proxy connections. */
			open=make_socks5_request("localhost", hostname, port, msg, verp, port); 
			if (open) break;

			/* FIXME */
			/* check user supplied hosts and ports for connection */

			if (config->testhost) {
				for (i=0;config->testhost[i];i++) {
					open=make_socks5_request(config->testhost[i]->hostname, hostname, port, msg, verp, config->testhost[i]->port); 
					if (open) break;
				}
			}
			break;
		default:
			IFDEBUG(print_debug(0, "Unknown Version: %i\n", ver));
			break;
	}
	free(msg);
	return open;
}

int make_socks5_request(const char *hostname, const char *sockhost, int port, char *msg, int *verp, int tport) {

	int r;
	int sock;
	char *resp, *res;
	char *msgp;

	char *results[] = { "Connection succeeded",
											"General SOCKS server failure",
											"Connection not allowed by ruleset",
											"Network unreachable",
											"Host unreachable",
											"Connection refused",
											"TTL expired",
											"Command not supported",
											"Address type not supported",
											"Unknown" };

	sock=net_connect(sockhost, port); /* we got this far so should connect */
	initial_request(sock, msg, verp); /* resent method */

	make_connect_request(msg, hostname, tport, *verp); /* make a connection string */
	net_nsend(sock, msg, 10); /* send it */

	resp=res=net_read(sock); /* get the socks response */
	res++;	/* ignore version byte (we know this is 5 already) */
	r=*res;	/* put connection result in an integer */
	free(resp); /* we don't need the result anymore */
	close(sock); /* close the connection - we don't need it */

	IFDEBUG(print_debug(1, "Result: %s\n", results[r<9?r:9]));

	if (r != 2) {
		IFDEBUG(print_debug(0, "Detected possible open socks server\n"));
		return 1;
	}
	return 0;
}

/* Socks 4 protocol...
 *
 *              +----+----+----+----+----+----+----+----+----+----+....+----+
 *              | VN | CD | DSTPORT |      DSTIP        | USERID       |NULL|
 *              +----+----+----+----+----+----+----+----+----+----+....+----+
 * # of bytes:	   1    1      2              4           variable       1
 *
 *
 * Responses:
 *              +----+----+----+----+----+----+----+----+
 *              | VN | CD | DSTPORT |      DSTIP        |
 *              +----+----+----+----+----+----+----+----+
 * # of bytes:	  1    1      2              4
 *
 * VN is the version of the reply code and should be 0. CD is the result
 * code with one of the following values:
 *
 * 90: request granted
 * 91: request rejected or failed
 * 92: request rejected because SOCKS server cannot connect to
 *     identd on the client
 * 93: request rejected because the client program and identd
 *     report different user-ids
 *
 *
 * Socks v5 protocol is described in rfc1928
 */

int make_socks4_request(const char *hostname, const char *sockhost, int port, char *msg, int *verp, int tport) {

	int r;
	int sock;
	char *resp, *res;
	char *msgp;

	char *results[] = { "Request succeeded",
											"Request rejected or failed",
											"Request rejected becasue SOCKS server cannot connect to identd on the client",
											"Request rejected because the client program and identd report different user-ids",
											"Unknown" };

	sock=net_connect(sockhost, port); /* we got this far so should connect */
	// initial_request(sock, msg, verp); /* resent method */

	make_connect_request(msg, hostname, tport, *verp); /* make a connection string */
	net_nsend(sock, msg, 16); /* send it */

	resp=res=net_read(sock); /* get the socks response */
	res++;	/* ignore version byte (we know this is 4 already) */
	r=*res;	/* put connection result in an integer */
	free(resp); /* we don't need the result anymore */
	close(sock); /* close the connection - we don't need it */

	if (!r) return(0);

	r-=90;

	if (r<0) r=4;

	IFDEBUG(print_debug(1, "Result: %s\n", results[r<4?r:4]));

	if (!r) {
		IFDEBUG(print_debug(1, "Detected possible open socks server\n"));
		return 1;
	}
	return 0;
}

int initial_request(int sock, char *msg, int *verp) {

	char *t, *res, *resp;
	int r;

	t=msg;
	memset(msg, '\0', NET_BUFFER); /* clear it */
	*t++=*verp; /* Version */
	*t++=3; /* Number of methods */
	*t++=0; /* Method 1 (No Auth) */
	*t++=1; /* Method 2 (GSAPI) */
	*t++=2; /* Method 3 (User/Pass) */
	net_nsend(sock, msg, 5); /* Send Auth Check/Request */
	resp=res=net_read(sock);
	
	*verp=*res++;
	
	switch (*verp) {
		case 0:
		case 4: /* kludge to ensure we catch broken socks v4 proxies */
			IFDEBUG(print_debug(3, "Socks Version 4 Reply\n"));
			break;
		case 5:
			IFDEBUG(print_debug(3, "Socks Version 5 Reply\n"));
			r=*res;
			if (r==255) {
				IFDEBUG(print_debug(1, "No Methods Accepted\n"));
			}
			break;
		default:
			IFDEBUG(print_debug(1, "Unknown Socks Version: %i\n", *(res-1)));
			break;
	}
	free(resp);
	return r;

}

int make_connect_request(char *msg, const char *hostname, int remp,  int ver) {

	int t;
	unsigned char *i;

	memset(msg, '\0', NET_BUFFER);
	*msg++=ver; /* Version */
	*msg++=1; /* Connect command */
	if (ver==5) {
		*msg++=0; /* Must set - Reserved */
		*msg++=1; /* Address type = IP Addy */
	} else {
		*msg++=remp >> 8; /* Port Number High Octet (right shift 8 places) */
		*msg++=remp & 0x00FF; /* Port Number Low Octet */
	}

	i=getip(hostname);
	IFDEBUG(printf("Debug[1]: ip="));
	for(t=0; t<4; t++, i++, msg++) {
		*msg=*i;
		IFDEBUG(printf("%hhi%s", *i, t<3?".":""));
	}
	IFDEBUG(printf("\n"));

	if (ver==5) {
		*msg++=remp >> 8; /* Port Number High Octet (right shift 8 places) */
		*msg++=remp & 0x00FF; /* Port Number Low Octet */
	} else {
		IFDEBUG(print_debug(3, "Setting UID to %s\n", userinfo->pw_name));
		strcpy(msg, userinfo->pw_name);
	}

}

