/* Debug Header */
#if DEBUG
# define IFDEBUG(x) do { if (debug) x; }while(0)
# define IFDEBUG5(x) do { if (debug) x; }while(0)
#else
# define IFDEBUG(x)
# define IFDEBUG5(x)
#endif
/* End Debug Header */


#if HAVE_CONFIG_H
# include <config.h>
#endif

#include "config.h"

#include <stdio.h>
#include <stdarg.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>

#include "extern.h"
#include "configuration.h"
#include "defines.h"

char *smtp_request(char tf, int type, char *ip, char *user, char *localdom, char *remotehost) {

	char *p, *s, *m;

	static char *mailfrom[] = {
/* 0  */		"MAIL FROM: <%u@%r>\r\n", 					/* spamtest@<remote domain> */
/* 1  */		"MAIL FROM: <%U@%d>\r\n", 					/* nobody@<local domain> */
/* 2  */		"MAIL FROM: <%u@nulhost.%d>\r\n",		/* spamtest@<non-existant host> */
/* 3  */		"MAIL FROM: <%u@localhost>\r\n",		/* spamtest@localhost */
/* 4  */		"MAIL FROM: <%u>\r\n", 							/* spamtest */
/* 5  */		"MAIL FROM: <>\r\n",			 					/* <> */
/* 6  */		"MAIL FROM: <%u@%r>\r\n", 					/* spamtest@<remote host> */
/* 7  */		"MAIL FROM: <%u@[%i]>\r\n", 				/* spamtest@[REM-IP-ADDRESS] */
/* 8  */		"MAIL FROM: <%u@%r>\r\n", 					/* spamtest@<remote host> */
/* 9  */		"MAIL FROM: <%u@%r>\r\n", 					/* spamtest@<remote host> */
/* 10 */		"MAIL FROM: <%u@%r>\r\n", 					/* spamtest@<remote host> */
/* 11 */		"MAIL FROM: <%u@%r>\r\n", 					/* spamtest@<remote host> */
/* 12 */		"MAIL FROM: <%u@[%i]>\r\n", 				/* spamtest@[REM-IP-ADDRESS] */
/* 13 */		"MAIL FROM: <%u@%r>\r\n", 					/* spamtest@<remote host> */
/* 14 */		"MAIL FROM: <%u@%r>\r\n", 					/* spamtest@<remote host> */
/* 15 */		"MAIL FROM: <%u@[%i]>\r\n", 				/* spamtest@[REM-IP-ADDRESS] */
/* 16 */		"MAIL FROM: <%u@%r>\r\n",		 				/* spamtest@<remote host> */
/* 17 */		"MAIL FROM: <%u@[%i]>\r\n",		 			/* spamtest@[REM-IP-ADDRESS] */
/* 18 */		"MAIL FROM: <%u@%r>\r\n",		 				/* spamtest@<remote host> */
/* 19 */		"MAIL FROM: <postmaster@%r>\r\n" };	/* postmaster@<remote host> */
	static char *rcptto[] = {
/* 0  */		"RCPT TO: <\"%U@%d\">\r\n", 				/* "nobody@<local domain>" */
/* 1  */		"RCPT TO: <%U@%d>\r\n", 						/* nobody@<local domain> */
/* 2  */		"RCPT TO: <%U@%d>\r\n", 						/* nobody@<local domain> */
/* 3  */		"RCPT TO: <%U@%d>\r\n", 						/* nobody@<local domain> */
/* 4  */		"RCPT TO: <%U@%d>\r\n", 						/* nobody@<local domain> */
/* 5  */		"RCPT TO: <%U@%d>\r\n", 						/* nobody@<local domain> */
/* 6  */		"RCPT TO: <%U@%d>\r\n", 						/* nobody@<local domain> */
/* 7  */		"RCPT TO: <%U@%d>\r\n", 						/* nobody@<local domain> */
/* 8  */		"RCPT TO: <%U%%%d@%r>\r\n",		 			/* nobody%<local domain>@<remote host> */
/* 9  */		"RCPT TO: <%U%%%d@[%i]>\r\n",		 		/* nobody%<local domain>@[REM-IP-ADDRESS] */
/* 10 */		"RCPT TO: <\"%U@%d\">\r\n", 				/* "nobody@<local domain>" */
/* 11 */		"RCPT TO: <\"%U%%%d\">\r\n", 				/* "nobody%<local domain>" */
/* 12 */		"RCPT TO: <%U@%d@%r\">\r\n", 				/* nobody@<local domain>@<remote host>" */
/* 13 */		"RCPT TO: <\"%U@%d\"@[%i]>\r\n",		/* "nobody@<local domain>"@[REM-IP-ADDRESS] */
/* 14 */		"RCPT TO: <@%r:%U@[%i]>\r\n",		 		/* nobody@<local domain>@[REM-IP-ADDRESS] */
/* 15 */		"RCPT TO: <@%r:%U@%d>\r\n", 					/* @<remote host>:nobody@<local domain> */
/* 16 */		"RCPT TO: <@[%i]:%U@%d>\r\n",				/* @[REM-IP-ADDRESS]:nobody@<local domain> */
/* 17 */		"RCPT TO: <%d!%U>\r\n",						 	/* <local domain>!nobody */
/* 18 */		"RCPT TO: <%d!%U@[%i]>\r\n",				/* <local domain>!nobody@[REM-IP-ADDRESS] */
/* 19 */		"RCPT TO: <%U@%d>\r\n" };			 			/* nobody@<local domain> */
	
	char buf[256] = "";

	switch(tf) {
		case 'f':
			s=strdup(mailfrom[type]);
			break;
		case 't':
			s=strdup(rcptto[type]);
			break;
		default:
			fprintf(stderr, "Oops, wrong type, someone screwed up... (smtp_request(tf))\n");
			s=malloc(1); /* safety will give a valid but empty string */
			*s='\0';
	}
	m=s;
	
	while(*s) {
		p=s;
		p=(char *)strchr(p, '%');
		if (!p) break;
		*p='\0';
		if (s) strcat(buf, s);
		p++;
		s=p;
		switch(*s++) {
			case 'U':
				strcat(buf, "spamtest");
				break;
			case 'd':
				strcat(buf, localdom);
				break;
			case 'r':
				strcat(buf, remotehost);
				break;
			case 'u':
				strcat(buf, user);
				break;
			case 'i':
				strcat(buf, ip);
				break;
			case '%':
				strcat(buf, "%");
			default:
				fprintf(stderr, "Oops, wrong type, someone screwed up... (smtp_request(%c))\n", *(s-1));
				break;
		}
	}
	if (s) strcat(buf, s);
		
	free(m);
	return(strdup(buf));
}

int test_relay (char *hostname) {

	char *buf, *mp;
	char msg[256];
	char tmsg[256];
	char mydomain[256];
	char myhost[256];
	static char rset[] = "RSET\r\n"; 
	int sockfd;
	int res, *resp;
	int ret;
	int i, c;
	char tf, ip[]="123.234.32.21";

	if (gethostname(myhost,256)) sprintf(myhost, "nowhere");
	if (getdomainname(mydomain,256)) sprintf(mydomain, "example.com");


	res=0;
	resp=&res;

	sockfd=net_connect(hostname, 25);

	if (!sockfd && !config->level) { /* Paranoid mode - reject if there is no remote smtp server */
		IFDEBUG(print_debug(4, "Rejecting SMTP as no connection to incoming server\n"));
		close(sockfd);
		if (config->mode==STANDALONE) reject(1);
		else config->thread->result=1;
		return 1;
	}

	if (!sockfd) return 0;

	mp=buf=net_read(sockfd);

	if (!buf) return 0;

	IFDEBUG(print_debug(4, "Server returned: %s\n", buf));
	
	res=(*buf=='2');

	while (res && buf && *buf=='2' && *(buf+3)=='-') {
		free(mp);
		mp=buf=net_read(sockfd);
		IFDEBUG(print_debug(4, "Server returned: %s\n", buf));
	}
	free(mp);

	sprintf(msg, "HELO %s\r\n", myhost);
	res=do_smtp_proto(sockfd, msg, '2');
	if (!res && config->strictsmtp) {
		ret=1;
		goto quit;
	}

	for (i=0;i<20;i++) {
		char *m;
		m=smtp_request('f', i, ip, config->spamtrapuser[1], config->spamtrapdom[1], hostname); /* FIXME - can only handle 1 spamtrap address */
		res=do_smtp_proto(sockfd, m, '2');
		if (!res) printf("Got bad result from SMTP server: %i (continuing...)\n", res);
		free(m);
		m=smtp_request('t', i, ip, config->spamtrapuser[1], config->spamtrapdom[1], hostname); /* FIXME - can only handle 1 spamtrap address */
		res=do_smtp_proto(sockfd, m, '2');
		free(m);
		if (!res) {
			printf("Got bad result from SMTP server: %i (RSETing and continuing...)\n", res);
			do_smtp_proto(sockfd, rset, '2');
		// continue;
		} else break; /* we got a 2xx response - server is likely open */
	}

quit:

	sprintf(msg, "QUIT\r\n");
	res=do_smtp_proto(sockfd, msg, '2');
	close(sockfd);

	return ret;
}

int do_smtp_proto(int sockfd, char *msg, int ok) {

	char *mp, *buf;
	char c, *t;
	int res;

	res=0;

	net_send(sockfd, msg);
	mp=buf=net_read(sockfd);
	IFDEBUG(print_debug(4, "Server returned: %s\n", buf));
	if (!buf) {
		IFDEBUG(print_debug(3, "Got NULL result (server probably dropped our connection)\n"));
		return 1;
	}
	res=(*buf==ok);
	*(buf+4) = '\0';
	if (!res) {
		IFDEBUG(print_debug(3, "Got bad result from SMTP server: %s\n", buf));
	} else {
		res=atoi(buf);
	}
	free(mp);
	return res;
}

/* do_send_mail():
 *
 * This function is to send mail to all the listed spamtrap addresses.
 * The only argument is an open socket descriptor, and the function
 * will not return any value, as it either sends the message or it 
 * doesn't - we don't care as the reception of the mail will list or not
 * the server in the spamtrap lists.
 */

void do_send_mail(int socket, char *buf) {

	typedef struct mail_proto mail_proto;
	struct mail_proto {
		char msg[128];
		int resp;
	};
			
	mail_proto protocol[] = {
		"MAIL FROM: %s\r\n", 2,
		"RCPT TO: %s\r\n", 2,
		"DATA\r\n", 3,
		"From: %s\r\nTo: %s\r\nSubject: Automated SORBS test - see: http://www.sorbs.net/\r\n\r\n%s\r\n.\r\n", 2,
		"QUIT\r\n", 2
	};

	if (get_server_helo(socket, buf) != 220) return;

	
	
	
}

int get_server_helo(int socket, char *buf) {

	char *p;
	int resp=0;
	char *junk;

	p=buf;

	while(*p && !isdigit(*p)) p++; /* skip to the first digit (if it exists) */
	if (!*p) { /* if it doesn't exist, get some more */
		free(buf);
		p=buf=net_read(socket);
		while(*p && !isdigit(*p)) p++; /* skip to the first digit (if it exists) */
	}
	while(*p && *(p+3)== '-') {
		while(*p && *p != '\r' && *p != '\n') p++;
		while(*p && !isdigit(*p)) p++;
	}
	if (!*p) return -1;
	scanf(p , "%i %s", &resp, junk);
		
	free(buf);
	return(resp); /* all ok */
}
